/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <iostream>

#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>

#include "ModelSolution.hh"

ModelSolution::ModelSolution(const ModelSpec &modspec_arg) throw (UnsupportedSpec)
  : modspec(modspec_arg)
{
}

ModelSolution::~ModelSolution()
{
}

void
ModelSolution::policy_func(const gsl_vector *y_prev, const gsl_vector *shocks, gsl_vector *y_curr)
{
  assert((int) y_prev->size == modspec.ny && (int) shocks->size == modspec.nx && (int) y_curr->size == modspec.ny);
  policy_func_internal(y_prev, shocks, y_curr);
}

void
ModelSolution::simulate_model(gsl_vector *init, gsl_matrix *sims, gsl_matrix *shocks,
                              int warmup, unsigned long rng_seed)
{
  const int T = sims->size1;
  assert((int) shocks->size1 == T && (int) shocks->size2 == modspec.nx
         && (int) init->size == modspec.ny && (int) sims->size2 == modspec.ny);

  // Initialize Mersenne Twister random number generator
  gsl_rng *rng = gsl_rng_alloc(gsl_rng_mt19937);
  gsl_rng_set(rng, rng_seed);

  // Simulate the initialization period, by updating init
  gsl_vector *y_tmp = gsl_vector_alloc(modspec.ny);
  gsl_vector *shocks_tmp = gsl_vector_alloc(modspec.nx);
  while(warmup--)
    {
      for(int j = 0; j < modspec.nx; j++)
        gsl_vector_set(shocks_tmp, j, gsl_ran_ugaussian(rng));

      policy_func(init, shocks_tmp, y_tmp);
      gsl_vector_memcpy(init, y_tmp);
    }
  gsl_vector_free(shocks_tmp);
  gsl_vector_free(y_tmp);

  // Simulate what we're interested in
  for(int t = 0; t < T; t++)
    {
      for(int j = 0; j < modspec.nx; j++)
        gsl_matrix_set(shocks, t, j, gsl_ran_ugaussian(rng));

      policy_func(t == 0 ? init : &gsl_matrix_const_row(sims, t-1).vector,
                  &gsl_matrix_const_row(shocks, t).vector,
                  &gsl_matrix_row(sims, t).vector);
    }

  gsl_rng_free(rng);
}
