/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cassert>
#include <iostream>
#include <sstream>

#include "MultiCountryRBCA2.hh"

MultiCountryRBCA2::MultiCountryRBCA2(int n_arg, double alpha_arg, double beta_arg, double delta_arg, double rho_arg, double sigma_arg, double phi_arg,
                                     double gamma_min_arg, double gamma_max_arg, double eta_min_arg, double eta_max_arg)
  : ModelSpec(5*n_arg+1, n_arg+1, 2*n_arg, n_arg),
    n(n_arg),
    alpha(alpha_arg), beta(beta_arg), delta(delta_arg), rho(rho_arg),
    sigma(sigma_arg), phi(phi_arg), A((1-beta)/(alpha*beta))
{
  taus = new double[n];
  gammas = new double[n];
  etas = new double[n];
  b = new double[n];
  for(int j = 0; j < n; j++)
    {
      gammas[j] = weighted_mass(gamma_min_arg, gamma_max_arg, j);
      etas[j] = weighted_mass(eta_min_arg, eta_max_arg, j);
      b[j] = (1-alpha)*pow(A, 1-1/gammas[j]);
      taus[j] = pow(A, 1/gammas[j]);
    }
}

MultiCountryRBCA2::~MultiCountryRBCA2()
{
  delete[] taus;
  delete[] gammas;
  delete[] etas;
  delete[] b;
}

double
MultiCountryRBCA2::weighted_mass(double min, double max, int j) const
{
  double w = ((double) j)/(((double) n) - 1);
  return (1-w)*min + w*max;
}

void
MultiCountryRBCA2::steady_state(gsl_vector *y_ss) const
{
  for(int j = 0; j < n; j++)
    {
      gsl_vector_set(y_ss, j, A);
      gsl_vector_set(y_ss, n+j, 1);
      gsl_vector_set(y_ss, 2*n+j, delta);
      gsl_vector_set(y_ss, 3*n+j, 1);
      gsl_vector_set(y_ss, 4*n+j, 1);
    }
  gsl_vector_set(y_ss, 5*n, 1);
}

void
MultiCountryRBCA2::errors(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *fwd, const gsl_vector *shocks, gsl_vector *errors) const
{
  assert(y_prev->stride == 1 && y_curr->stride == 1 && errors->stride == 1);
  assert((int) y_prev->size == 5*n+1 && (int) y_curr->size == 5*n+1 && (int) fwd->size == n
         && (int) shocks->size == n+1 && (int) errors->size == 5*n+1);

  const double *c_curr = y_curr->data;
  const double *i_curr = y_curr->data + 2*n;
  const double *k_curr = y_curr->data + 3*n;
  const double *a_curr = y_curr->data + 4*n;
  const double lambda_curr = gsl_vector_get(y_curr, 5*n);

  const double *k_prev = y_prev->data + 3*n;
  const double *a_prev = y_prev->data + 4*n;

  double uc_curr[n], ul_curr[n], f_curr[n], fk_curr[n], fl_curr[n];

  marginal_utilities(y_curr->data, uc_curr, ul_curr);
  production(y_prev->data, y_curr->data, f_curr, fk_curr, fl_curr);

  double *errs = errors->data;

  for(int j = 0; j < n; j++)
    {
      errs[j] = (taus[j] * uc_curr[j] - lambda_curr)/(taus[j] * uc_curr[j]);
      errs[n+j] = (taus[j] * ul_curr[j] + lambda_curr * a_curr[j] * fl_curr[j]) / (taus[j] * ul_curr[j]);
      double lhs = lambda_curr * (1 + phi * (i_curr[j] / k_prev[j] - delta));
      errs[2*n+j] = (lhs - beta * gsl_vector_get(fwd, j))/lhs;
      errs[3*n+j] = (k_curr[j] - i_curr[j] - (1 - delta) * k_prev[j]) / k_curr[j];
      errs[4*n+j] = (a_curr[j] - exp(rho * log(a_prev[j]) + sigma * (gsl_vector_get(shocks, j) + gsl_vector_get(shocks, n)))) / a_curr[j];
    }

  double lhs = 0, rhs = 0;
  for(int j = 0; j < n; j++)
    {
      lhs += c_curr[j] + i_curr[j] - delta * k_prev[j];
      rhs += a_curr[j] * f_curr[j] - phi / 2 * k_prev[j] * pow(i_curr[j] / k_prev[j] - delta, 2);
    }
  errs[5*n] = 1 - rhs / lhs;
}

void
MultiCountryRBCA2::forward_part(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *y_next, const gsl_vector *shocks, gsl_vector *fwd) const
{
  assert(y_prev->stride == 1 && y_curr->stride == 1 && y_next->stride == 1);
  assert((int) y_prev->size == 5*n+1 && (int) y_curr->size == 5*n+1
         && (int) y_next->size == 5*n+1 && (int) fwd->size == n
         && (int) shocks->size == n+1);

  const double *k_curr = y_curr->data + 3*n;

  const double *i_next = y_next->data + 2*n;
  const double *a_next = y_next->data + 4*n;
  const double lambda_next = gsl_vector_get(y_next, 5*n);

  double f_next[n], fk_next[n], fl_next[n];
  production(y_curr->data, y_next->data, f_next, fk_next, fl_next);

  for(int j = 0; j < n; j++)
    gsl_vector_set(fwd, j, lambda_next * (1.0 + a_next[j] * fk_next[j] + phi * (1.0 + 0.5 * (i_next[j] / k_curr[j] - delta)) * (i_next[j] / k_curr[j] - delta)));
}

void
MultiCountryRBCA2::write_common_params(std::ostream &output) const
{
  output << "n=" << n << ", alpha=" << alpha << ", beta=" << beta
         << ", delta=" << delta << ", sigma=" << sigma
         << ", rho=" << rho << ", phi=" << phi << ", ";
}

void
MultiCountryRBCA2::write_state(std::ostream &output, const gsl_vector *y) const
{
  int i;
  output << "Consumption:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(y, i);
  output << std::endl
         << "Labor:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(y, i+n);
  output << std::endl
         << "Investment:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(y, i+2*n);
  output << std::endl
         << "Capital:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(y, i+3*n);
  output << std::endl
         << "Technology:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(y, i+4*n);
  output << std::endl
         << "Lambda: " << gsl_vector_get(y, 5*n) << std::endl;
}

void
MultiCountryRBCA2::write_errors(std::ostream &output, const gsl_vector *errs) const
{
  int i;
  output << "Marginal utility of consumption:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(errs, i);
  output << std::endl
         << "Marginal utility of labor:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(errs, i+n);
  output << std::endl
         << "Euler equation:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(errs, i+2*n);
  output << std::endl
         << "Law of motion of capital:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(errs, i+3*n);
  output << std::endl
         << "Law of motion of technology:";
  for(i = 0; i < n; i++)
    output << " " << gsl_vector_get(errs, i+4*n);
  output << std::endl
         << "World budget constraint: " << gsl_vector_get(errs, 5*n) << std::endl;
}


void
MultiCountryRBCA2::production(const double y_prev[], const double y_curr[], double f[], double fk[], double fl[]) const
{
  const double *k_prev = y_prev + 3*n;
  const double *l_curr = y_curr + n;
  for(int j = 0; j < n; j++)
    {
      f[j] = A * pow(k_prev[j], alpha) * pow(l_curr[j], 1 - alpha);
      fk[j] = alpha * A * pow(k_prev[j] / l_curr[j], alpha - 1);
      fl[j] = (1 - alpha) * A * pow(k_prev[j] / l_curr[j], alpha);
    }
}

void
MultiCountryRBCA2::marginal_utilities(const double y_curr[], double uc[], double ul[]) const
{
  const double *c_curr = y_curr;
  const double *l_curr = y_curr + n;
  for(int j = 0; j < n; j++)
    {
      uc[j] = pow(c_curr[j], -1/gammas[j]);
      ul[j] = -b[j]*pow(l_curr[j], 1/etas[j]);
    }
}

void
MultiCountryRBCA2::write_output(std::ostream &output) const
{
  output << "RBC multi-country asymetric no. 2: ";
  write_common_params(output);
  output << "gamma=[ ";
  for(int j = 0; j < n; j++)
    output << gammas[j] << " ";
  output << "], eta=[ ";
  for(int j = 0; j < n; j++)
    output << etas[j] << " ";
  output << "]" << std::endl;
}

void
MultiCountryRBCA2::states_names(std::vector<std::string> &r) const
{
  std::ostringstream s;
  
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "k_prev_" << i;
      r.push_back(s.str());
    }
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "a_" << i;
      r.push_back(s.str());
    }
}

void
MultiCountryRBCA2::shocks_names(std::vector<std::string> &r) const
{
  std::ostringstream s;
  
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "e_" << i;
      r.push_back(s.str());
    }

  r.push_back("e_global");
}

void
MultiCountryRBCA2::variables_names(std::vector<std::string> &r) const
{
  std::ostringstream s;
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "c_" << i;
      r.push_back(s.str());
    }

  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "l_" << i;
      r.push_back(s.str());
    }

  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "i_" << i;
      r.push_back(s.str());
    }

  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "k_" << i;
      r.push_back(s.str());
    }

  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "a_" << i;
      r.push_back(s.str());
    }

  r.push_back("lambda");
}

void
MultiCountryRBCA2::residuals_names(std::vector<std::string> &r) const
{
  std::ostringstream s;
  
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "MargConsUtil_" << i;
      r.push_back(s.str());
    }
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "MargLabUtil_" << i;
      r.push_back(s.str());
    }
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "Euler_" << i;
      r.push_back(s.str());
    }
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "LawMotionCap_" << i;
      r.push_back(s.str());
    }
  for(int i = 1; i <= n; i++)
    {
      s.str("");
      s << "TFP_" << i;
      r.push_back(s.str());
    }
  r.push_back("AggRscConst");
}

std::string
MultiCountryRBCA2::name() const
{
  std::ostringstream s;
  s << "MC-RBC-A2-N" << n;
  return s.str();
}

void
MultiCountryRBCA2::state_steady_state(gsl_vector *s_ss) const
{
  gsl_vector_set_all(s_ss, 1);
}

void
MultiCountryRBCA2::infer_from_state(const gsl_vector *state, gsl_vector *y_prev, gsl_vector *shocks) const
{
  /* Infer y_prev and shocks from this state by supposing that a_t is the result
     of a_{t-1}=1 and null global shock */
  gsl_vector_set_all(y_prev, NAN);
  gsl_vector_set(shocks, n, 0);
  for(int i = 0; i < n; i++)
    {
      gsl_vector_set(y_prev, 3*n+i, gsl_vector_get(state, i));
      gsl_vector_set(y_prev, 4*n+i, 1);
      gsl_vector_set(shocks, i, log(gsl_vector_get(state, n+i))/sigma);
    }
}

void
MultiCountryRBCA2::extract_state(const gsl_vector *init, const gsl_matrix *sims, gsl_matrix *state) const
{
  assert(sims->size2 == (size_t) ny && init->size == (size_t) ny && state->size2 == (size_t) ns);
  assert(sims->size1 == state->size1);
  const int T = (int) sims->size1;

  // Copy a_t
  gsl_matrix_memcpy(&gsl_matrix_submatrix(state, 0, 0, T, n).matrix,
                    &gsl_matrix_const_submatrix(sims, 0, 4*n, T, n).matrix);
  // Copy k_{t-1}
  gsl_vector_memcpy(&gsl_matrix_subrow(state, 0, n, n).vector,
                    &gsl_vector_const_subvector(init, 3*n, n).vector);
  gsl_matrix_memcpy(&gsl_matrix_submatrix(state, 1, n, T-1, n).matrix,
                    &gsl_matrix_const_submatrix(sims, 0, 3*n, T-1, n).matrix);

}
