/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _MULTI_COUNTRY_RBC_A2_HH
#define _MULTI_COUNTRY_RBC_A2_HH

#include "ModelSpec.hh"

//! Multi country RBC: asymetric specicification no. 2 in JEDC 2011 comparison project
/*
  Country indices j go from 0 to n-1

  Variables is a vector of size 5*n+1:
  - c^j at j
  - l^j at n+j
  - i^j at 2*n+j
  - k^j at 3*n+j
  - a^j at 4*n+j
  - lambda at 5*n

  Shocks is a vector of size n+1:
  - e^j at j
  - e at n

  Errors is a vector of size 5*n+1.
*/
class MultiCountryRBCA2 : public ModelSpec
{
  friend class MultiCountryCGASolution;
  friend class MultiCountryHybSolution;
private:
  double *b, *gammas, *etas, *taus;
  void write_common_params(std::ostream &output) const;
  double weighted_mass(double min, double max, int j) const;
  void marginal_utilities(const double y_curr[], double uc[], double ul[]) const;
  /*! We need previous and current variables because production is done using capital of previous period and labor of current period */
  void production(const double y_prev[], const double y_curr[], double f[], double fk[], double fl[]) const;
public:
  //! Number of countries
  const int n;
  const double alpha, beta, delta, rho, sigma, phi, A;

  MultiCountryRBCA2(int n_arg, double alpha_arg, double beta_arg, double delta_arg, double rho_arg, double sigma_arg, double phi_arg, double gamma_min_arg, double gamma_max_arg, double eta_min_arg, double eta_max_arg);
  virtual ~MultiCountryRBCA2();
  virtual void steady_state(gsl_vector *y_ss) const;
  virtual void errors(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *fwd, const gsl_vector *shocks, gsl_vector *errors) const;
  virtual void forward_part(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *y_next, const gsl_vector *shocks, gsl_vector *fwd) const;
  virtual void state_steady_state(gsl_vector *s_ss) const;
  virtual void infer_from_state(const gsl_vector *state, gsl_vector *y_prev, gsl_vector *shocks) const;
  virtual void extract_state(const gsl_vector *init, const gsl_matrix *sims, gsl_matrix *state) const;
  virtual void write_output(std::ostream &output) const;
  virtual void write_state(std::ostream &output, const gsl_vector *y) const;
  virtual void write_errors(std::ostream &output, const gsl_vector *errs) const;
  virtual void states_names(std::vector<std::string> &r) const;
  virtual void shocks_names(std::vector<std::string> &r) const;
  virtual void variables_names(std::vector<std::string> &r) const;
  virtual void residuals_names(std::vector<std::string> &r) const;
  virtual std::string name() const;
};

#endif // !_MULTI_COUNTRY_RBC_A2_HH
