/*
 * Copyright (C) 2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cassert>
#include <cmath>

#include <gsl/gsl_blas.h>

#include <matio.h>

#include "PerSolution.hh"

#define CHECK_ARRAY(x, m, n) assert((x) != NULL \
                                    && x->class_type == MAT_C_DOUBLE \
                                    && x->rank == 2  \
                                    && x->dims[0] == (m)     \
                                    && x->dims[1] == (n))

PerSolution::PerSolution(const ModelSpec &modspec_arg, const std::string &matname, int order_arg) throw (UnsupportedSpec)
  : ModelSolution(modspec_arg), ny(modspec_arg.ny), nz(modspec_arg.ns+modspec_arg.nx),
    order(order_arg)
{
  assert(order >= 1 && order <= 3);

  // Retrieve MATLAB arrays
  mat_t *mfp = Mat_Open(matname.c_str(), MAT_ACC_RDONLY);
  if (mfp == NULL)
    {
      std::cerr << "PerSolution::PerSolution: can't open " << matname << std::endl;
      throw UnsupportedSpec();
    }

  matvar_t *ys_mx = Mat_VarReadInfo(mfp, (char *) "dyn_ss");
  matvar_t *g_0_mx = Mat_VarReadInfo(mfp, (char *) "dyn_g_0");
  matvar_t *g_1_mx = Mat_VarReadInfo(mfp, (char *) "dyn_g_1");

  matvar_t *g_2_mx = NULL, *g_3_mx = NULL;

  if (order >= 2)
    g_2_mx = Mat_VarReadInfo(mfp, (char *) "dyn_g_2");
  if (order >= 3)
    g_3_mx = Mat_VarReadInfo(mfp, (char *) "dyn_g_3");

  // Test array types and dimensions
  CHECK_ARRAY(ys_mx, ny, 1);
  Mat_VarReadDataAll(mfp, ys_mx);
  CHECK_ARRAY(g_0_mx, ny, 1);
  Mat_VarReadDataAll(mfp, g_0_mx);
  CHECK_ARRAY(g_1_mx, ny, nz);
  Mat_VarReadDataAll(mfp, g_1_mx);
  if (order >= 2)
    {
      CHECK_ARRAY(g_2_mx, ny, nz*(nz+1)/2);
      Mat_VarReadDataAll(mfp, g_2_mx);
    }
  if (order >= 3)
    {
      CHECK_ARRAY(g_3_mx, ny, nz*(nz+1)*(nz+2)/6);
      Mat_VarReadDataAll(mfp, g_3_mx);
    }

  // Convert to GSL vectors and matrices
  ys = gsl_vector_alloc(ny);
  gsl_vector_memcpy(ys, &gsl_vector_view_array((double *) ys_mx->data, ny).vector);

  g_0 = gsl_vector_alloc(ny);
  gsl_vector_memcpy(g_0, &gsl_vector_view_array((double *) g_0_mx->data, ny).vector);

  const gsl_matrix *g_1_prime = &gsl_matrix_view_array((double *) g_1_mx->data, nz, ny).matrix;
  g_1 = gsl_matrix_alloc(ny, nz);
  gsl_matrix_transpose_memcpy(g_1, g_1_prime);

  if (order >= 2)
    {
      const gsl_matrix *g_2_prime = &gsl_matrix_view_array((double *) g_2_mx->data, nz*(nz+1)/2, ny).matrix;
      g_2 = gsl_matrix_alloc(ny, nz*(nz+1)/2);
      gsl_matrix_transpose_memcpy(g_2, g_2_prime);
    }

  if (order >= 3)
    {
      const gsl_matrix *g_3_prime = &gsl_matrix_view_array((double *) g_3_mx->data, nz*(nz+1)*(nz+2)/6, ny).matrix;
      g_3 = gsl_matrix_alloc(ny, nz*(nz+1)*(nz+2)/6);
      gsl_matrix_transpose_memcpy(g_3, g_3_prime);
    }

  // Free MX arrays
  Mat_VarFree(ys_mx);
  Mat_VarFree(g_0_mx);
  Mat_VarFree(g_1_mx);
  if (order >= 2)
    Mat_VarFree(g_2_mx);
  if (order >= 3)
    Mat_VarFree(g_3_mx);
  Mat_Close(mfp);

  // Allocate temporary variables
  state = gsl_vector_alloc(nz);
  y_reordered = gsl_vector_alloc(ny);
  if (order >= 2)
    state2 = gsl_vector_alloc(nz*(nz+1)/2);
  if (order >= 3)
    state3 = gsl_vector_alloc(nz*(nz+1)*(nz+2)/6);
}

PerSolution::~PerSolution()
{
  gsl_vector_free(ys);
  gsl_vector_free(g_0);
  gsl_matrix_free(g_1);
  if (order >= 2)
    gsl_matrix_free(g_2);
  if (order >= 3)
    gsl_matrix_free(g_3);
  gsl_vector_free(state);
  gsl_vector_free(y_reordered);
  if (order >= 2)
    gsl_vector_free(state2);
  if (order >= 3)
    gsl_vector_free(state3);
}

void
PerSolution::policy_func_perturbation(const gsl_vector *y_prev, const gsl_vector *shocks, gsl_vector *y_curr)
{
  create_state_vector(y_prev, shocks, state);

  // Set to the steady state
  gsl_vector_memcpy(y_reordered, ys);

  // Add constant correction
  gsl_vector_add(y_reordered, g_0);

  // Compute first order terms
  gsl_blas_dgemv(CblasNoTrans, 1.0, g_1, state, 1.0, y_reordered);

  // Compute second order terms
  if (order >= 2)
    {
      size_t j = 0;
      for (size_t i0 = 0; i0 < (size_t) nz; i0++)
        for (size_t i1 = i0; i1 < (size_t) nz; i1++)
          {
            double mult = (i0 == i1) ? 1 : 2;
            gsl_vector_set(state2, j, gsl_vector_get(state, i0)*gsl_vector_get(state, i1)*mult);
            j++;
          }
      gsl_blas_dgemv(CblasNoTrans, 1.0, g_2, state2, 1.0, y_reordered);
    }

  // Compute third order terms
  if (order >= 3)
    {
      size_t j = 0;
      for (size_t i0 = 0; i0 < (size_t) nz; i0++)
        for (size_t i1 = i0; i1 < (size_t) nz; i1++)
          for (size_t i2 = i1; i2 < (size_t) nz; i2++)
            {
              double mult;
              if (i0 == i1 && i1 == i2)
                mult = 1;
              else if (i0 == i1 || i1 == i2 || i0 == i2)
                mult = 3;
              else
                mult = 6;
              gsl_vector_set(state3, j, gsl_vector_get(state, i0)*gsl_vector_get(state, i1)
                             *gsl_vector_get(state, i2)*mult);
              j++;
          }
      gsl_blas_dgemv(CblasNoTrans, 1.0, g_3, state3, 1.0, y_reordered);
    }

  reorder_from_dynare(y_reordered, y_curr);
}
