/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cmath>
#include <climits>
#include <cassert>
#include <fstream>
#include <sstream>
#include <iomanip>

#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_qrng.h>

#include "SolutionTester.hh"

SolutionTester::SolutionTester(ModelSolution &modsol_arg, integration_method int_meth_arg)
  : modsol(modsol_arg),
    ny(modsol.modspec.ny), nx(modsol.modspec.nx),
    ns(modsol.modspec.ns), nf(modsol.modspec.nf),
    int_meth(int_meth_arg),
    monomial_deg5_d(nx),
    monomial_deg5_r(sqrt(1.0+0.5*monomial_deg5_d)),
    monomial_deg5_s(sqrt(0.5+0.25*monomial_deg5_d)),
    monomial_deg5_V(pow(M_PI, monomial_deg5_d/2.0)),
    monomial_deg5_A(2.0/(monomial_deg5_d+2.0)*monomial_deg5_V),
    monomial_deg5_B((4.0-monomial_deg5_d)/(2.0*pow(monomial_deg5_d+2.0, 2.0))*monomial_deg5_V),
    monomial_deg5_D(monomial_deg5_V/pow(monomial_deg5_d+2.0, 2.0))
{
}

// Values from http://mathworld.wolfram.com/Hermite-GaussQuadrature.html
const double SolutionTester::hermite_nodes[HERMITE_NODES_NB] = {
  sqrt((3.0-sqrt(6.0))/2.0),
  -sqrt((3.0-sqrt(6.0))/2.0),
  sqrt((3.0+sqrt(6.0))/2.0),
  -sqrt((3.0+sqrt(6.0))/2.0)
};

const double SolutionTester::hermite_weights[HERMITE_NODES_NB] = {
  sqrt(M_PI)/(4.0*(3.0-sqrt(6.0))),
  sqrt(M_PI)/(4.0*(3.0-sqrt(6.0))),
  sqrt(M_PI)/(4.0*(3.0+sqrt(6.0))),
  sqrt(M_PI)/(4.0*(3.0+sqrt(6.0)))
};

void
SolutionTester::euler_errors_gauss_hermite(const gsl_vector *y_prev, const gsl_vector *y_curr,
                                           const gsl_vector *shocks_curr, gsl_vector *errors)
{
  assert(y_prev->size == (size_t) ny);

  // Compute the expectancy of the forward part of the Euler equation
  gsl_vector *fwd_mean = gsl_vector_alloc(nf);
  gsl_vector_set_zero(fwd_mean);

  gsl_vector *shocks_next = gsl_vector_alloc(nx); 
  gsl_vector *y_next = gsl_vector_alloc(ny);
  gsl_vector *fwd_point = gsl_vector_alloc(nf);

  assert(pow(HERMITE_NODES_NB, nx) <= ((double) INT_MAX));

  /* Iteration over the grid in n+1 dimensions is done using an integer
     which is then interpreted as n+1 digits in base HERMITE_NODES_NB */
  int point = 0;
  while(point < pow(HERMITE_NODES_NB, nx))
    {
      double weight = 1;
      int idx = point;
      for(int i = 0; i < nx; i++)
        {
          gsl_vector_set(shocks_next, i, hermite_nodes[idx % HERMITE_NODES_NB]);
          weight *= hermite_weights[idx % HERMITE_NODES_NB] * M_2_SQRTPI / 2.0;
          idx /= HERMITE_NODES_NB;
        }

      gsl_vector_scale(shocks_next, M_SQRT2);
      modsol.policy_func(y_curr, shocks_next, y_next);
      modsol.modspec.forward_part(y_prev, y_curr, y_next, shocks_curr, fwd_point);
      gsl_vector_scale(fwd_point, weight);
      gsl_vector_add(fwd_mean, fwd_point);

      point++;
    }

  // Compute Euler errors for all equations
  modsol.modspec.errors(y_prev, y_curr, fwd_mean, shocks_curr, errors);

  gsl_vector_free(fwd_point);
  gsl_vector_free(y_next);
  gsl_vector_free(shocks_next);
  gsl_vector_free(fwd_mean);
}

void
SolutionTester::euler_errors_monomial_deg5(const gsl_vector *y_prev, const gsl_vector *y_curr,
                                           const gsl_vector *shocks_curr, gsl_vector *errors)
{
  assert(y_prev->size == (size_t) ny);

  // Compute the expectancy of the forward part of the Euler equation
  gsl_vector *fwd_mean = gsl_vector_alloc(nf);
  gsl_vector_set_zero(fwd_mean);

  gsl_vector *shocks_next = gsl_vector_alloc(nx); 
  gsl_vector *y_next = gsl_vector_alloc(ny);
  gsl_vector *fwd_point = gsl_vector_alloc(nf);

  // First part of formula (term at origin)
  gsl_vector_set_zero(shocks_next);
  modsol.policy_func(y_curr, shocks_next, y_next);
  modsol.modspec.forward_part(y_prev, y_curr, y_next, shocks_curr, fwd_point);
  gsl_vector_scale(fwd_point, monomial_deg5_A);
  gsl_vector_add(fwd_mean, fwd_point);

  // Second part of formula
  // Index s is 0 or 1, for choosing the sign of the point
  for(int i = 0; i < nx; i++)
    for(int s = 0; s < 2; s++)
      {
        gsl_vector_set_zero(shocks_next);
        if (s)
          gsl_vector_set(shocks_next, i, M_SQRT2 * monomial_deg5_r);
        else
          gsl_vector_set(shocks_next, i, -M_SQRT2 * monomial_deg5_r);
        modsol.policy_func(y_curr, shocks_next, y_next);
        modsol.modspec.forward_part(y_prev, y_curr, y_next, shocks_curr, fwd_point);
        gsl_vector_scale(fwd_point, monomial_deg5_B);
        gsl_vector_add(fwd_mean, fwd_point);
      }

  // Third part of formula
  // Index s is between 0 and 3: bit 0 determines sign for index i, bit 1 determines sign for index j
  for(int i = 0; i < nx-1; i++)
    for(int j = i+1; j < nx; j++)
      for(int s = 0; s < 4; s++)
        {
          gsl_vector_set_zero(shocks_next);

          if (s & 1)
            gsl_vector_set(shocks_next, i, M_SQRT2 * monomial_deg5_s);
          else
            gsl_vector_set(shocks_next, i, -M_SQRT2 * monomial_deg5_s);

          if (s & 2)
            gsl_vector_set(shocks_next, j, M_SQRT2 * monomial_deg5_s);
          else
            gsl_vector_set(shocks_next, j, -M_SQRT2 * monomial_deg5_s);

          modsol.policy_func(y_curr, shocks_next, y_next);
          modsol.modspec.forward_part(y_prev, y_curr, y_next, shocks_curr, fwd_point);
          gsl_vector_scale(fwd_point, monomial_deg5_D);
          gsl_vector_add(fwd_mean, fwd_point);
        }

  // Rescale by 1/sqrt(PI)^d
  gsl_vector_scale(fwd_mean, pow(M_PI, -monomial_deg5_d/2.0));

  // Compute Euler errors for all equations
  modsol.modspec.errors(y_prev, y_curr, fwd_mean, shocks_curr, errors);

  gsl_vector_free(fwd_point);
  gsl_vector_free(y_next);
  gsl_vector_free(shocks_next);
  gsl_vector_free(fwd_mean);
}

void
SolutionTester::euler_errors_quasi_monte_carlo(const gsl_vector *y_prev, const gsl_vector *y_curr,
                                               const gsl_vector *shocks_curr, gsl_vector *errors)
{
  assert(y_prev->size == (size_t) ny);

  // Compute the expectancy of the forward part of the Euler equation
  gsl_vector *fwd_mean = gsl_vector_alloc(nf);
  gsl_vector_set_zero(fwd_mean);

  gsl_vector *shocks_next = gsl_vector_alloc(nx); 
  gsl_vector *y_next = gsl_vector_alloc(ny);
  gsl_vector *fwd_point = gsl_vector_alloc(nf);

  // Use Niederreiter low-discrepancy sequences
  if (nx > 12)
    {
      std::cerr << "SolutionTester::euler_errors_quasi_monte_carlo: too high dimensionality for Niederreiter generator" << std::endl;
      exit(EXIT_FAILURE);
    }
  gsl_qrng *qrng = gsl_qrng_alloc(gsl_qrng_niederreiter_2, nx);

  for(int i = 0; i < QMC_POINTS_NB; i++)
    {
      gaussian_draw(shocks_next, qrng);
      modsol.policy_func(y_curr, shocks_next, y_next);
      modsol.modspec.forward_part(y_prev, y_curr, y_next, shocks_curr, fwd_point);
      gsl_vector_add(fwd_mean, fwd_point);
    }
  gsl_vector_scale(fwd_mean, 1.0/((double) QMC_POINTS_NB));

  // Compute Euler errors for all equations
  modsol.modspec.errors(y_prev, y_curr, fwd_mean, shocks_curr, errors);

  gsl_qrng_free(qrng);
  gsl_vector_free(fwd_point);
  gsl_vector_free(y_next);
  gsl_vector_free(shocks_next);
  gsl_vector_free(fwd_mean);
}

void
SolutionTester::euler_errors(const gsl_vector *y_prev, const gsl_vector *y_curr,
                             const gsl_vector *shocks_curr, gsl_vector *errors)
{
  assert(y_prev->size == y_curr->size && y_prev->size == errors->size
         && (int) shocks_curr->size == nx);

#ifdef DEBUG
  // Compare GH and Monomial methods
  gsl_vector *errs2 = gsl_vector_alloc(ny);
  euler_errors_gauss_hermite(y_prev, y_curr, shocks_curr, errors);
  euler_errors_monomial_deg5(y_prev, y_curr, shocks_curr, errs2);
  gsl_vector_sub(errs2, errors);
  std::cout << "Diff:" << std::endl;
  gsl_vector_fprintf(stdout, errs2, "%.10g");
  std::cout << std::endl;
  gsl_vector_free(errs2);
#endif

  switch(int_meth)
    {
    case GaussHermiteAndMonomialDegree5:
      if (nx <= MAX_GAUSS_HERMITE_DIM)
        euler_errors_gauss_hermite(y_prev, y_curr, shocks_curr, errors);
      else
        euler_errors_monomial_deg5(y_prev, y_curr, shocks_curr, errors);
      break;
    case GaussHermiteOnly:
      euler_errors_gauss_hermite(y_prev, y_curr, shocks_curr, errors);
      break;
    case MonomialDegree5Only:
      euler_errors_monomial_deg5(y_prev, y_curr, shocks_curr, errors);
      break;
    case QuasiMonteCarloOnly:
      euler_errors_quasi_monte_carlo(y_prev, y_curr, shocks_curr, errors);
      break;
    }
}

void
SolutionTester::gaussian_draw(gsl_vector *x, gsl_qrng *qrng)
{
  assert(x->size == qrng->dimension);

  double *rnd = new double[x->size];

  /* Draw a quasi-random sample in the [0,1] hypercube, excluding those on the edges
     and convert it to a gaussian vector using inverse cumulative */
 loop:
  gsl_qrng_get(qrng, rnd);
  for(int j = 0; j < (int) x->size; j++)
    {
      if (rnd[j] == 0.0 || rnd[j] == 1.0)
        goto loop;
      gsl_vector_set(x, j, gsl_cdf_ugaussian_Pinv(rnd[j]));
    }
  delete[] rnd;
}

void
SolutionTester::hypersphere_draw(gsl_vector *x, const gsl_vector *center, 
                                 double radius, gsl_qrng *qrng)
{
  // Draw a non-zero random gaussian vector
  do
    gaussian_draw(x, qrng);
  while(gsl_blas_dnrm2(x) == 0.0);

  gsl_vector_scale(x, radius / gsl_blas_dnrm2(x));
  gsl_vector_add(x, center);
}


void
SolutionTester::accuracy_test1(int nb_points, double radius, double &max_err, gsl_matrix *data)
{
  assert(data == NULL || (data->size1 == (size_t) nb_points && data->size2 == (size_t) (ns+2*ny)));

  gsl_vector *state_ss = gsl_vector_alloc(ns);
  gsl_vector *y_prev = gsl_vector_alloc(ny);
  gsl_vector *y_curr = gsl_vector_alloc(ny);
  gsl_vector *shocks = gsl_vector_alloc(nx);
  gsl_vector *errors = gsl_vector_alloc(ny);

  // The state space has 2*n dimensions, for the variables k_{t-1} and a_t
  gsl_vector *state = gsl_vector_alloc(ns);

  modsol.modspec.state_steady_state(state_ss);

  // Initialize max errors to NAN
  max_err = NAN;

  /* Use Sobol low-discrepancy sequences (Niederreiter sequences are only valid up
     to 12 dimensions, which is not enough for 8 or 10 countries */
  assert(ns <= 40);

  gsl_qrng *qrng = gsl_qrng_alloc(gsl_qrng_sobol, ns);

  for(int i = 0; i < nb_points; i++)
    {
      hypersphere_draw(state, state_ss, radius, qrng);

      modsol.modspec.infer_from_state(state, y_prev, shocks);

      // Compute policy function and Euler error
      modsol.policy_func(y_prev, shocks, y_curr);
      euler_errors(y_prev, y_curr, shocks, errors);

      for(int j = 0; j < ny; j++)
        {
          double e = log10(fabs(gsl_vector_get(errors, j)));

          if (std::isnan(max_err) || e > max_err)
            max_err = e;
        }

      if (data != NULL)
        {
          // Write results
          for(int j = 0; j < ns; j++)
            gsl_matrix_set(data, i, j, gsl_vector_get(state, j));
          for(int j = 0; j < ny; j++)
            gsl_matrix_set(data, i, ns+j, gsl_vector_get(y_curr, j));
          for(int j = 0; j < ny; j++)
            gsl_matrix_set(data, i, ns+ny+j, gsl_vector_get(errors, j));
        }
    }

  gsl_vector_free(state);
  gsl_vector_free(errors);
  gsl_vector_free(shocks);
  gsl_vector_free(y_curr);
  gsl_vector_free(y_prev);
  gsl_vector_free(state_ss);
  gsl_qrng_free(qrng);
}

void
SolutionTester::accuracy_test2(int T, int warmup, double &max_err, double &mean_err,
                               unsigned long rng_seed, gsl_matrix *data)
{
  assert(data == NULL || (data->size1 == (size_t) T && data->size2 == (size_t) (ny+nx+ny)));

  gsl_vector *init = gsl_vector_alloc(ny);

  // Use steady state as starting point for simulations
  modsol.modspec.steady_state(init);

#ifdef DEBUG
  std::cout << "Computing simulations..." << std::endl;
#endif

  gsl_matrix *sims = gsl_matrix_alloc(T, ny);
  gsl_matrix *shocks = gsl_matrix_alloc(T, nx);
  modsol.simulate_model(init, sims, shocks, warmup, rng_seed);
  
  max_err = NAN;
  mean_err = 0;

  gsl_vector *errors = gsl_vector_alloc(ny);
  for(int t = 0; t < T; t++)
    {
      euler_errors(t == 0 ? init : &gsl_matrix_row(sims, t-1).vector,
                   &gsl_matrix_row(sims, t).vector,
                   &gsl_matrix_row(shocks, t).vector,
                   errors);

      for(int j = 0; j < ny; j++)
        {
          double ee = fabs(gsl_vector_get(errors, j));
          mean_err += ee;

          double e = log10(ee);
          
          if (std::isnan(max_err) || e > max_err)
            max_err = e;
        }

      if (data != NULL)
        {
          // Write results
          for(int j = 0; j < ny; j++)
            gsl_matrix_set(data, t, j, gsl_matrix_get(sims, t, j));
          for(int j = 0; j < nx; j++)
            gsl_matrix_set(data, t, ny+j, gsl_matrix_get(shocks, t, j));
          for(int j = 0; j < ny; j++)
            gsl_matrix_set(data, t, ny+nx+j, gsl_vector_get(errors, j));
        }
    }

  mean_err = log10(mean_err / (T*ny));

  gsl_vector_free(errors);
  gsl_matrix_free(sims);
  gsl_matrix_free(shocks);
  gsl_vector_free(init);
}

#ifdef DEBUG
void
SolutionTester::print_matrix(const gsl_matrix *A)
{
  for(int i = 0; i < (int) A->size1; i++)
    {
      for(int j = 0; j < (int) A->size2; j++)
        std::cout << gsl_matrix_get(A, i, j) << " ";
      std::cout << std::endl;
    }
  std::cout << std::endl;
}

void
SolutionTester::hypersphere_draw_test(int dim, double zmin, double zmax, int ndraw)
{
  gsl_vector *center = gsl_vector_alloc(dim);
  gsl_vector_set_zero(center);

  gsl_vector *x = gsl_vector_alloc(dim);

  gsl_qrng *qrng = gsl_qrng_alloc(gsl_qrng_sobol, dim);

  int *counts = new int[dim];
  for(int i = 0; i < dim; i++)
    counts[i] = 0;

  for(int i = 0; i < ndraw; i++)
    {
      hypersphere_draw(x, center, 1, qrng);
      for(int j = 0; j < dim; j++)
        if (gsl_vector_get(x, j) >= zmin && gsl_vector_get(x, j) <= zmax)
          counts[j]++;
    }

  std::cout << "Counts: ";
  for(int i = 0; i < dim; i++)
    std::cout << counts[i] << " ";
  std::cout << std::endl;

  delete[] counts;
  gsl_qrng_free(qrng);
  gsl_vector_free(x);
  gsl_vector_free(center);
}
#endif
