/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vector>
#include <ostream>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_qrng.h>

#include "MultiCountryRBCA2.hh"
#include "ModelSolution.hh"

class SolutionTester
{
public:
  enum integration_method
    {
      GaussHermiteAndMonomialDegree5,
      GaussHermiteOnly,
      MonomialDegree5Only,
      QuasiMonteCarloOnly
    };

private:
  ModelSolution &modsol;
  //! Copied from modsol
  const int ny, nx, ns, nf;
  //! Numerical integration method to be used
  const integration_method int_meth;

  //! Thrown when matrix is not inversible in prewithen() or in inv_newey_west_estimator()
  class SingularityException
  {
  };

  //! Number of nodes used for Gauss-Hermite integration
  const static int HERMITE_NODES_NB = 4;
  //! Nodes for Gauss-Hermite integration (roots of Hermite polynomials)
  static const double hermite_nodes[HERMITE_NODES_NB];
  //! Weights for Gauss-Hermite integration
  static const double hermite_weights[HERMITE_NODES_NB];
  //! Computes Euler errors using product Gauss-Hermite integration
  /*! \todo Should value functions be sorted before adding them ? */
  void euler_errors_gauss_hermite(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *shocks_curr, gsl_vector *errors);

  //! Constants used in monomial degree 5 integration formula
  const double monomial_deg5_d, monomial_deg5_r, monomial_deg5_s, monomial_deg5_V, monomial_deg5_A, monomial_deg5_B, monomial_deg5_D;
  //! Computes Euler errors using monomial degree 5 integration formula
  void euler_errors_monomial_deg5(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *shocks_curr, gsl_vector *errors);

  //! Number of points for Quasi Monte-Carlo integration
  const static int QMC_POINTS_NB = 1000;
  //! Computes Euler errors using Quasi Monte-Carlo integration (low discrepancy sequence)
  /*! \todo Should value functions be sorted before adding them ? */
  void euler_errors_quasi_monte_carlo(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *shocks_curr, gsl_vector *errors);

  //! Dimensionality above which we switch to Quasi Monte-Carlo
  const static int MAX_GAUSS_HERMITE_DIM = 6;
  //! Computes Euler errors, selecting between possible numerical methods
  void euler_errors(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *shocks_curr, gsl_vector *errors);

#ifdef DEBUG
  static void print_matrix(const gsl_matrix *A);
#endif

  //! Quasi-randomly draw a multivariate standard normal vector
  static void gaussian_draw(gsl_vector *x, gsl_qrng *qrng);
  //! Quasi-randomly draw a point on a hypersphere
  /*!
    \param x the random draw
    \param center the hypersphere center
    \param radius the hypersphere radius
    \param qrng quasi-random number generator
  */
  static void hypersphere_draw(gsl_vector *x, const gsl_vector *center, double radius, gsl_qrng *qrng);

public:
  SolutionTester(ModelSolution &modsol_arg, integration_method int_meth_arg);
  //! Accuracy test 1: Euler equation errors on points of a hypersphere around steady state
  /*!
    \param nb_points number of points drawn on the sphere
    \param radius radius of the sphere
    \param max_err on output, maximum error over all points and all equations (in base 10 log)
    \param data if not NULL, on output, a matrix containing nb_points rows, and
                in columns: state variables, endogenous variables, residuals (in levels)
  */
  void accuracy_test1(int nb_points, double radius, double &max_err, gsl_matrix *data = NULL);
  //! Accuracy test 2: Dynamic Euler equations errors
  /*!
    \param T number of periods for each simulation
    \param warmup number of periods to discard at the beginning of each simulation
    \param max_err on output, maximum error over all periods and all equations (in base 10 log)
    \param mean_err on output, mean error over all periods and all equations (in base 10 log)
    \param rng_seed a seed for the random generator
    \param data if not NULL, on output, a matrix containing T rows, and
                in columns: endogenous variables, exogenous variables, residuals (in levels)
  */
  void accuracy_test2(int T, int warmup, double &max_err, double &mean_err, unsigned long rng_seed, gsl_matrix *data = NULL);
#ifdef DEBUG
  static void hypersphere_draw_test(int dim, double zmin, double zmax, int ndraw);
#endif
};
