/*
 * Copyright (C) 2008-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cstring>
#include <unistd.h>
#include <iomanip>
#include <iostream>
#include <vector>

#include <sys/time.h>
#include <sys/resource.h>

#include "MultiCountryRBCA2.hh"
#include "SolutionTester.hh"

#include "MultiCountryCGASolution.hh"
#include "MultiCountryHybSolution.hh"

double
timeval_substract(const struct timeval &x, const struct timeval &y)
{
  return ((double) (x.tv_sec - y.tv_sec))
    + ((double) (x.tv_usec - y.tv_usec)) / 1000000.0;
}

double
cputime(const struct rusage &t1, const struct rusage &t2)
{
  return round(100.0*(timeval_substract(t2.ru_utime, t1.ru_utime)
                    + timeval_substract(t2.ru_stime, t1.ru_stime)))/100.0;
}

int
main(int argc, char** argv)
{
  int opt;
  int begin_spec = 1, end_spec = 0, test_no = 0;
  int N1 = 1000;  // default nbr of points for Test 1
  int N2 = 10000; // default nbr of points for Test 2
  int N3 = 10000; // default nbr of points for Test 3
  char *selected_solution = NULL;
  unsigned long rng_seed = 0;
  SolutionTester::integration_method int_meth = SolutionTester::GaussHermiteAndMonomialDegree5;

  while((opt = getopt(argc, argv, "b:B:gmqr:s:t:u:v:w:")) != -1)
    {
      switch(opt)
        {
        case 'b':
          begin_spec = atoi(optarg);
          break;
        case 'B':
          begin_spec = atoi(optarg);
          end_spec = begin_spec;
          break;
        case 'g':
          int_meth = SolutionTester::GaussHermiteOnly;
          break;
        case 'm':
          int_meth = SolutionTester::MonomialDegree5Only;
          break;
        case 'q':
          int_meth = SolutionTester::QuasiMonteCarloOnly;
          break;
        case 'r':
          rng_seed = atoi(optarg);
          break;
        case 's':
          selected_solution = optarg;
          break;
        case 't':
          test_no = atoi(optarg);
          break;
        case 'u':
          N1 = atoi(optarg);
          break;
        case 'v':
          N2 = atoi(optarg);
          break;
        case 'w':
          N3 = atoi(optarg);
          break;
        default:
          std::cerr << "Usage: " << argv[0] << " [-b INTEGER] [-B INTEGER] [-d] [-g] [-m] [-q] [-r INTEGER] [-s NAME] [-t INTEGER] [-u INTEGER] [-v INTEGER] [-w INTEGER]" << std::endl;
          exit(EXIT_FAILURE);
        }
    }

  double alpha = 0.36, beta = 0.99, delta = 0.025, sigma = 0.01, rho = 0.95, phi = 0.5;

  std::vector<MultiCountryRBCA2 *> models;

  // Construct models
  int sizes[] = { 2, 4, 6, 8, 10, 20, 40, 60, 80, 100 };
  for(int i = 0; i < 10; i++)
    models.push_back(new MultiCountryRBCA2(sizes[i], alpha, beta, delta, rho, sigma, phi, 0.25, 1, 0.1, 1));

  if (begin_spec < 1 || begin_spec > (int) models.size())
    {
      std::cerr << "Invalid specification number " << begin_spec << std::endl;
      exit(EXIT_FAILURE);
    }

  if (end_spec == 0)
    {
      end_spec = models.size();
    }

  if (end_spec < begin_spec || end_spec > (int) models.size())
    {
      std::cerr << "Invalid specification number " << begin_spec << std::endl;
      exit(EXIT_FAILURE);
    }

  std::cout << std::setprecision(10);

#ifdef DEBUG
  for(std::vector<MultiCountryRBCA2 *>::const_iterator it = models.begin();
      it != models.end(); ++it)
    {
      // Display model specification
      (*it)->write_output(std::cout);

      // Fetch steady state and display it
      const int n = (*it)->n;
      gsl_vector *y_ss = gsl_vector_alloc(5*n+1);
      (*it)->steady_state(y_ss);
      std::cout << "Steady state:" << std::endl;
      (*it)->write_state(std::cout, y_ss);

      // Compute max theoretical error at steady state
      gsl_vector *shocks = gsl_vector_alloc(n+1);
      gsl_vector_set_zero(shocks);
      gsl_vector *fwd = gsl_vector_alloc(n);
      (*it)->forward_part(y_ss, y_ss, y_ss, shocks, fwd);
      double max_err = (*it)->max_error(y_ss, y_ss, fwd, shocks);
      std::cout << "Max theoretical error at steady state: " << max_err << std::endl;

      // Compute policy functions errors at steady state with zero shock
      gsl_vector *y_curr = gsl_vector_alloc(5*n+1);
      gsl_vector *y_next = gsl_vector_alloc(5*n+1);

      // Determine the solution methods to be tested
      std::vector<ModelSolution *> solutions;
      solutions.push_back(new MultiCountryCGASolution(**it));

      for(std::vector<ModelSolution *>::iterator it2 = solutions.begin();
          it2 != solutions.end(); ++it2)
        {
          (*it2)->policy_func(y_ss, shocks, y_curr);
          (*it2)->policy_func(y_curr, shocks, y_next);

          // Display max error at steady state
          (*it)->forward_part(y_ss, y_curr, y_next, shocks, fwd);
          double max_err = (*it)->max_error(y_ss, y_curr, fwd, shocks);
          std::cout << "Max error for " << (*it2)->name() << " at steady state with zero shock: " << max_err << std::endl;

          delete *it2;
        }

      std::cout << std::endl;

      gsl_vector_free(shocks);
      gsl_vector_free(fwd);
      gsl_vector_free(y_curr);
      gsl_vector_free(y_next);
      gsl_vector_free(y_ss);
    }
#endif

  gsl_vector *thresholds = gsl_vector_alloc(3);
  gsl_vector_set(thresholds, 0, 0.05);
  gsl_vector_set(thresholds, 1, 0.5);
  gsl_vector_set(thresholds, 2, 0.95);

  // Perform tests
  for(std::vector<MultiCountryRBCA2 *>::const_iterator it = models.begin() + begin_spec - 1;
      it != models.begin() + end_spec; ++it)
    {
      // Display model specification
      std::cout << "*** ";
      (*it)->write_output(std::cout);
      std::cout << std::endl;

      // Determine the solution methods to be tested
      std::vector<ModelSolution *> solutions;
      if (selected_solution == NULL)
        {
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 1, MultiCountryHybSolution::PlainPerturbation));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 1, MultiCountryHybSolution::Hybrid));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 2, MultiCountryHybSolution::PlainPerturbation));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 2, MultiCountryHybSolution::Hybrid));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 3, MultiCountryHybSolution::PlainPerturbation));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
          try
            {
              solutions.push_back(new MultiCountryHybSolution(**it, 3, MultiCountryHybSolution::Hybrid));
            }
          catch(ModelSolution::UnsupportedSpec &e)
            {
            }
        }
      else
        {
          if (!strcmp(selected_solution, "cga"))
            solutions.push_back(new MultiCountryCGASolution(**it));
          else if (!strcmp(selected_solution, "hyb1"))
            solutions.push_back(new MultiCountryHybSolution(**it, 1, MultiCountryHybSolution::Hybrid));
          else if (!strcmp(selected_solution, "hyb2"))
            solutions.push_back(new MultiCountryHybSolution(**it, 2, MultiCountryHybSolution::Hybrid));
          else if (!strcmp(selected_solution, "hyb3"))
            solutions.push_back(new MultiCountryHybSolution(**it, 3, MultiCountryHybSolution::Hybrid));
          else if (!strcmp(selected_solution, "per1"))
            solutions.push_back(new MultiCountryHybSolution(**it, 1, MultiCountryHybSolution::PlainPerturbation));
          else if (!strcmp(selected_solution, "per2"))
            solutions.push_back(new MultiCountryHybSolution(**it, 2, MultiCountryHybSolution::PlainPerturbation));
          else if (!strcmp(selected_solution, "per3"))
            solutions.push_back(new MultiCountryHybSolution(**it, 3, MultiCountryHybSolution::PlainPerturbation));
          else
            {
              std::cerr << "Unknown solution method " << selected_solution << std::endl;
              exit(EXIT_FAILURE);
            }
        }
      
      for(std::vector<ModelSolution *>::iterator it2 = solutions.begin();
          it2 != solutions.end(); ++it2)
        {
          SolutionTester st(**it2, int_meth);

          double max_err;
          if (!test_no || test_no == 1)
            {
              struct rusage r1, r2, r3, r4;
              std::cout << "Accuracy test 1 on " << (*it2)->name() << ":" << std::endl;
              std::cout << " Nbr of points " << N1 << std::endl;
              std::cout << " Radius 0.01: ";
              getrusage(RUSAGE_SELF, &r1);
              st.accuracy_test1(N1, 0.01, max_err);
              getrusage(RUSAGE_SELF, &r2);
              std::cout << max_err << " (cputime: " << cputime(r1, r2) << "s)" << std::endl;
              std::cout << " Radius 0.10: ";
              st.accuracy_test1(N1, 0.1, max_err);
              getrusage(RUSAGE_SELF, &r3);
              std::cout << max_err << " (cputime: " << cputime(r2, r3) << "s)"  << std::endl;
              std::cout << " Radius 0.30: ";
              st.accuracy_test1(N1, 0.3, max_err);
              getrusage(RUSAGE_SELF, &r4);
              std::cout << max_err << " (cputime: " << cputime(r3, r4) << "s)"  << std::endl;
              std::cout << std::endl;
            }

          double mean_err;
          if (!test_no || test_no == 2)
            {
              struct rusage r1, r2;
              std::cout << "Accuracy test 2 on " << (*it2)->name() << ":" << std::endl;
              std::cout << " Nbr of points " << N2 << std::endl;
              getrusage(RUSAGE_SELF, &r1);
              st.accuracy_test2(N2, 200, max_err, mean_err, rng_seed);
              getrusage(RUSAGE_SELF, &r2);
              std::cout << " Max:  " << max_err << std::endl
                        << " Mean: " << mean_err << std::endl
                        << " (cputime: " << cputime(r1, r2) << "s)" << std::endl
                        << std::endl;
            }
          
          delete *it2;
        }

      std::cout << std::endl;

    }

  // Cleanup
  for(std::vector<MultiCountryRBCA2 *>::iterator it = models.begin();
      it != models.end(); ++it)
    delete (*it);
  gsl_vector_free(thresholds);

#ifdef DEBUG
  SolutionTester::hypersphere_draw_test(16, 0.2, 0.3, 100000);
#endif

  return EXIT_SUCCESS;
}
